#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       main.py
#       
#       Copyright 2009-2011 Sandro Mani <manisandro@gmail.com>
#       
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gtk
import os
import sys
import locale
import urllib2
import gettext
import webbrowser
import tempfile
import Image
import TiffImagePlugin	# Explicit import needed for py2exe
import subprocess

import acquire
import config
import dialogs
import drawingarea
import notifier
import threadtools
import outbuffer
import filelist
import enchant

# Fix the dict dir
if sys.platform[0:3]=="win":
	enchant.set_param("enchant.myspell.dictionary.path", os.environ['DICTDIR'])
	
class gimagereader:
	def __init__(self, pydir, files=None):
		# Internationalization
		if os.path.exists(os.path.join(pydir,'locale')):
			localedir=os.path.join(pydir,'locale')
		else:
			localedir=os.path.join(os.sep,'usr','share','locale')
		if os.getenv('LANG') is None:	# windows hack for locale setting
			lang, enc = locale.getdefaultlocale()
			os.environ['LANG'] = lang
		# Translation of _( ) strings
		gettext.bindtextdomain('gimagereader', localedir)
		gettext.textdomain('gimagereader')
		gettext.install('gimagereader', localedir)
		_ = gettext.gettext
		# Translation of gtkbuilder UI
		try:
			locale.setlocale(locale.LC_ALL, '')
		except:
			print "Failed to load GUI translations..."
		try:
			locale.bindtextdomain('gimagereader', localedir)
		except:	# windows hack for gtkbuilder translations
			try:
				import ctypes
				libintl = ctypes.cdll.LoadLibrary("intl.dll") 
				libintl.bindtextdomain('gimagereader', localedir)
			except:
				print "Failed to load GUI translations..."
			
		# Widgets
		self.builder=gtk.Builder()
		self.builder.set_translation_domain('gimagereader')
		self.builder.add_from_file(os.path.join(pydir,'gimagereader.xml'))
		self.window=self.builder.get_object("main")
		self.statusbar=self.builder.get_object("statusbar")
		self.aboutdialog=self.builder.get_object("aboutdialog")
		self.notebook_sources=self.builder.get_object("notebook_sources")
		self.layout_sources=self.builder.get_object("layout_sources")
		self.eventbox_sources_close=self.builder.get_object("eventbox_sources_close")
		self.menu_toggle_sources=self.builder.get_object("menu_toggle_sources")
		self.tb_doocr=self.builder.get_object("tb_doocr")
		self.tb_stopocr=self.builder.get_object("tb_stopocr")
		
		# Some variables
		self.normalstate=True
		self.statusctx=self.statusbar.get_context_id("main")
		self.statusstackheight=0
		self.layout_sources_size=[0, 0]
		self.lang=[None, None, None]
		self.pages=[]
		self.outtext=""
		self.cancelled=False

		# Complete the GUI
		self.langmenuhead=self.builder.get_object("tb_langmenu")
		self.langmenuhead.set_use_underline(False)
		self.langmenuhead.get_children()[0].get_children()[1].hide()
		
		# Load modules
		self.acquire=acquire.Acquire(self.builder, self)
		self.drawingarea=drawingarea.Drawingarea(self.builder, self)
		self.versionnotifier=notifier.VersionNotifier(self.builder)
		self.conf=config.Config(self.builder)
		self.outbuffer=outbuffer.Outbuffer(self.builder, self)
		self.filelist=filelist.FileList(self.builder, self)
		
		# Connect signals and show window
		self.window.connect("delete-event", self.quit)
		self.window.connect("window-state-event", self.__store_window_state)
		self.langmenuhead.connect("clicked", lambda w: self.langmenuhead.get_children()[0].get_children()[1].clicked())
		self.builder.get_object("menu_config").connect("activate", self.show_config)
		self.builder.get_object("menu_guide").connect("activate", lambda w: dialogs.help())
		self.builder.get_object("menu_about").connect("activate", self.__show_about)
		self.builder.get_object("menu_quit").connect("activate", self.quit)
		self.menu_toggle_sources.connect("toggled", self.__toggle_sources)
		self.layout_sources.connect("size-allocate", self.__layout_sources_size_changed)
		self.builder.get_object("button_sources_close").connect("clicked", self.__close_sources)
		self.tb_stopocr.connect("clicked", self.__recognize_stop)
		
		# Set geometry
		geo=self.conf.get_default_geometry()
		if geo!=None:
			self.window.resize(geo[2],geo[3])
			self.window.move(geo[0],geo[1])
		self.builder.get_object("hpaned_output").set_orientation(self.conf.get_pane_orientation())		
		self.window.show()
		
		# Start
		self.set_status(_("Select an image to begin..."))
		if not self.conf.get_valid():
			self.show_config()
		else:
			self.load_dictionaries()
		
		if self.conf.get_update_check()==True:
			threadtools.SubprocessThread(self.__getnewversion, self.__checkversion).start()
		
		if files!=None:
			self.filelist.open_images(None, files)
	
	### Public interface functions ###
	def quit(self,Widget=None,event=None):
		if not self.outbuffer.clear_buffer():
			return True
		self.outbuffer.detach_spellchecker()		# Prevent windows from crash-exiting
		if self.normalstate:
			self.conf.set_default_geometry(self.window.get_position(),self.window.get_size())
			self.conf.save_config()
		gtk.main_quit()
		sys.exit(0)
		
	def get_config(self):
		""" Returns the configuration instance. """
		return self.conf
	
	def get_drawingarea(self):
		""" Returns the drawingarea instance. """
		return self.drawingarea
	
	def get_filelist(self):
		""" Returns the filelist instance. """
		return self.filelist
	
	def push_status(self, message):
		""" Pushes a message on the statusbar. """
		self.statusbar.push(self.statusctx, message)
		self.statusstackheight+=1
	
	def pop_status(self):
		""" Pops the last message from the statusbar. """
		self.statusbar.pop(self.statusctx)
		self.statusstackheight-=1
	
	def set_status(self, message):
		""" Sets the status of the statusbar. Note that his clears the stack of existing messages. """
		for i in range(0, self.statusstackheight):
			self.statusbar.pop(self.statusctx)
		self.statusstackheight=1
		self.statusbar.push(self.statusctx, message)
	
	def set_busy(self, busy, message=None):
		""" Sets whether the the application is busy (wait cursor, insensitive window). If busy=True, the message argument is mandatory. """
		if self.builder.get_object("menubar").get_property('sensitive')!=busy:
			return False
		if busy:
			if message==None: print "Warning: received None message along with busy=True in main::set_busy"
			self.window.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))
			self.drawingarea.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))
			self.builder.get_object("menubar").set_sensitive(False)
			self.builder.get_object("vbox_body").set_sensitive(False)
			self.builder.get_object("toolbar").foreach(self.__set_tb_sensitive, False)
			self.statusbar.push(self.statusctx, message)
			self.statusstackheight+=1
		else:
			self.builder.get_object("menubar").set_sensitive(True)
			self.builder.get_object("vbox_body").set_sensitive(True)
			self.builder.get_object("toolbar").foreach(self.__set_tb_sensitive, True)
			self.window.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.LEFT_PTR))
			self.drawingarea.set_cursor(None)
			self.statusbar.pop(self.statusctx)
			self.statusstackheight-=1
		return False	# In case was called by a gobject.timeout_add
	
	def __set_tb_sensitive(self, widget, sensitive):
		if widget!=self.tb_stopocr:
			widget.set_sensitive(sensitive)
	
	### Sources Pane Functions ###
	def __toggle_sources(self, widget=None):
		if not (self.layout_sources.flags() & gtk.VISIBLE):
			self.layout_sources.show()
		else:
			self.layout_sources.hide()
	
	def __close_sources(self, widget):
		self.menu_toggle_sources.set_active(False)
	
	def __layout_sources_size_changed(self, widget, allocation):
		if allocation.width==self.layout_sources_size[0] and allocation.height==self.layout_sources_size[1]: return
		self.layout_sources_size=[allocation.width, allocation.height]
		self.notebook_sources.set_size_request(allocation.width, allocation.height)
		self.layout_sources.move(self.eventbox_sources_close, allocation.width-self.eventbox_sources_close.allocation.width, -1)
	
	### Recognition ###
	def recognize(self, pages=None):
		""" Recognizes the text in the range of pages. If pages is not specified, the current image/selection is recognized. """
		self.set_busy(True, _("Recognizing, please wait..."))
		self.tb_stopocr.show()
		self.tb_doocr.hide()
		self.tb_stopocr.set_sensitive(True)
		self.cancelled=False
		self.outtext=""
		if pages:
			self.pages=pages
			threadtools.SubprocessThread(self.__recognize_do, self.__recognize_done, self.drawingarea.get_selection(self.pages[0])).start()
		else:
			threadtools.SubprocessThread(self.__recognize_do, self.__recognize_done, self.drawingarea.get_selection()).start()
	
	def __recognize_do(self, image):
		tess=self.conf.get_paths(['tess'])[0]
		[temptifdesc,temptif]=tempfile.mkstemp('.tif','tmpocr')
		try:
			image.save(temptif,'TIFF',compression='raw')
		except:
			return [False, _("Failed to perform recognition"), _("Could not save selection to temporary folder.")]
		# Recognize
		[tmptxtdesc,tmptxt]=tempfile.mkstemp('.txt','tmpocr');
		p=subprocess.Popen([tess, temptif, tmptxt[:-4], "-l", self.lang[0]], shell=False, stderr=subprocess.PIPE, stdout=open(os.devnull,'a'), creationflags=threadtools.popencf)
		out=p.communicate()
		os.close(temptifdesc)
		os.remove(temptif)
		if p.returncode!=0 or not os.path.isfile(tmptxt):
			return [False, _("Failed to perform recognition"), "tesseract "+_("returned")+":\n"+out[1]+"\n"+_("This probabily is a bug in tesseract, retry using an image at a different resolution or by varying the selected region.")]
		# Read output
		outfile = open(tmptxt, "r")
		self.outtext+=outfile.read()
		os.close(tmptxtdesc)
		outfile.close()
		os.remove(tmptxt)
		# Set current / output filename
		return [True]
	
	def __recognize_done(self, result):
		""" Finalizes the recognition process, initiated by see self.recognize. """
		if self.cancelled:
			self.outtext=""
			return
		if result[0]!=True:
			dialogs.error_dialog(result[1], result[2], self.window)
		else:
			self.pages=self.pages[1:]
			if self.pages:
				threadtools.SubprocessThread(self.__recognize_do, self.__recognize_done, self.drawingarea.get_selection(self.pages[0])).start()
				return
			self.outbuffer.set_buffer(self.outtext, self.filelist.suggest_output())
		self.set_busy(False)
		self.tb_stopocr.hide()
		self.tb_doocr.show()
		self.outtext=""
	
	def __recognize_stop(self, widget):
		self.cancelled=True
		self.set_busy(False)
		self.tb_stopocr.hide()
		self.tb_doocr.show()
	
	### Configuration ###
	def __store_window_state(self, widget, event):
		self.normalstate=(event.new_window_state==0)
	
	def show_config(self ,widget=None):
		self.conf.show_dialog()
		if not self.conf.get_valid():
			self.window.connect('event-after', self.quit)
		else:
			self.load_dictionaries()
			self.builder.get_object("hpaned_output").set_orientation(self.conf.get_pane_orientation())
	
	def load_dictionaries(self):
		langmenu=gtk.Menu()
		alldicts=enchant.list_dicts()
		checkitem=None
		langset=False
		activelang=self.lang[2]
		if not activelang:
			activelang=self.conf.get_pref_lang()
		for lang in self.conf.get_languages():
			item=gtk.MenuItem(lang[1])
			submenu=gtk.Menu()
			dicts=[ d[0] for d in alldicts if d[0][0:2]==lang[2][0:2] ]
			dicts.sort()
			if not dicts: dicts=[lang[2]]
			for d in dicts:
				checkitem=gtk.RadioMenuItem(checkitem, d, False)
				checkitem.connect("activate", self.__change_language, [lang[0], lang[1], d])
				submenu.append(checkitem)
				if d==activelang:
					checkitem.set_active(True)
					langset=True
			item.set_submenu(submenu)
			langmenu.append(item)
		if not langset:
			langmenu.get_children()[0].get_submenu().get_children()[0].set_active(True)
		self.langmenuhead.set_menu(langmenu)
		langmenu.show_all()
	
	def __change_language(self, widget, lang):
		self.langmenuhead.set_label("%s (%s)"%(lang[1], lang[2]))
		self.outbuffer.set_language(lang)
		self.lang=lang
	
	### About ###
	def __show_about(self, widget):
		""" Callback to the about entry of the help menu. """
		self.aboutdialog.run()
		self.aboutdialog.hide()
	
	### Check for updates ###
	def __getnewversion(self):
		""" Retreives the latest published program version. """
		try:
			urlh = urllib2.urlopen("http://sourceforge.net/projects/gimagereader/files/LATEST/download?use_mirror=autoselect",None,20)
		except:
			return "unknown"
		return urlh.read()
	
	def __checkversion(self, txt):
		""" Compares the latest published program version (see self.__getnewversion) to the current program version. """
		if txt=="unknown":
			return
		nv=txt.replace('\n','').split('.')
		cv=self.aboutdialog.get_version()
		if cv[-4:]=="-svn":
			cv=cv[:-4]
		cv=cv.split('.')
		l=min(len(cv),len(nv))
		for i in range(l):	#Find first version point number where unequal and check if newer
			if int(nv[i])!=int(cv[i]):
				if int(nv[i])>int(cv[i]):
					self.versionnotifier.notify(_('New version'), 'gImageReader '+txt.replace('\n','')+' '+_('is available.'), (('Download', lambda w, e: webbrowser.open('http://sourceforge.net/projects/gimagereader/files')),('Changelog',lambda w, e: webbrowser.open('http://sourceforge.net/projects/gimagereader/files/changelog.txt/download?use_mirror=autoselect'))))
				break
	
