#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       outbuffer.py
#       
#       Copyright 2010-2011 Sandro Mani <manisandro@gmail.com>
#       
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This file contains the class for handling the output text buffer

import os
import gtk
import subprocess
import re
import sys
import gobject
import gettext
_ = gettext.gettext

import dialogs
import notifier
import threadtools

try:
	import gtkspell
except:
	gtkspell=None
	print >> sys.stderr, "Could not load gtkspell, spellcheck will not work."

if os.path.isfile('/usr/bin/apt-file'):
	haveaptfile=True
	havepackagekit=False
else:
	try:
		import dbus
		import dbus.mainloop.glib
		dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
		havepackagekit=dbus.SessionBus().get_object('org.freedesktop.PackageKit', '/org/freedesktop/PackageKit')
		havepackagekit=True
		haveaptfile=False
	except:
		haveaptfile=havepackagekit=False
		print >> sys.stderr, "Could not load dbus, find PackageKit on dbus or find apt-file: dictionary autoinstallation will not work."
	
class Outbuffer:
	""" The class controlling the output pane. """
	def __init__(self, builder, main):
		self.outputfile=os.path.join(os.path.expanduser("~"), "ocr_output.txt")
		self.spellchecker=None
		self.notifier=notifier.Notifier(builder)
		self.main=main
		self.append_mode=0
		self.lang=[None, None, None]
		self.searchiter=None
		
		self.window=builder.get_object("main")
		self.status=builder.get_object("status_msg")
		self.outputbox=builder.get_object("outputbox")
		self.hpane=builder.get_object("hpaned_output")
		self.output=builder.get_object("output")
		self.btn_stripcrlf=builder.get_object("btn_stripcrlf")
		self.menu_keepifdot=builder.get_object("menu_keepifdot")
		self.menu_keepifquest=builder.get_object("menu_keepifquest")
		self.menu_keepifexcl=builder.get_object("menu_keepifexcl")
		self.menu_keepifquot=builder.get_object("menu_keepifquot")
		self.findreplacebox=builder.get_object("findreplacebox")
		self.entry_find=builder.get_object("entry_find")
		self.entry_replace=builder.get_object("entry_replace")
		
		# Complete GUI
		self.outbuffer=gtk.TextBuffer(None)
		self.output.set_buffer(self.outbuffer)
		self.appendcombo = gtk.combo_box_new_text()
		builder.get_object("tb_output_append_mode").add(self.appendcombo)
		self.appendcombo.append_text(_("At cursor"))
		self.appendcombo.append_text(_("At end"))
		self.appendcombo.append_text(_("Replace"))
		self.appendcombo.set_active(0)
		self.appendcombo.show()
		self.btn_stripcrlf.set_sensitive(False)
		self.findreplacebox.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse("#FFE090"))
		
		# Callbacks
		builder.get_object("btn_stripcrlf").connect("clicked",self.__stripcrlf)
		builder.get_object("btn_save").connect("clicked",self.save_buffer)
		builder.get_object("btn_clearout").connect("clicked",self.clear_buffer)
		self.outbuffer.connect("notify::has-selection",self.__onselect)
		self.outbuffer.connect("notify::cursor-position", self.__set_search_iter)
		builder.get_object("btn_replace").connect("clicked", self.__toggle_findreplace)
		builder.get_object("btn_find").connect("clicked", self.__find_in_buffer)
		self.entry_find.connect("activate", self.__find_in_buffer)
		self.entry_find.connect("changed", self.__entry_find_changed)
		builder.get_object("btn_findreplace").connect("clicked", self.__find_in_buffer, True)
		self.entry_replace.connect("activate", self.__find_in_buffer, True)
		self.appendcombo.connect("changed", self.__append_mode_changed)
	
	### Language functions ###
	def set_language(self, lang):
		""" Sets the language for the output buffer, used for setting spellchecking. """
		self.lang=lang
		if not gtkspell:
			return
		if self.spellchecker!=None:
			gtkspell.Spell.detach(self.spellchecker)
			self.spellchecker=None
		try:
			self.spellchecker=gtkspell.Spell(self.output,lang[2])
			self.notifier.hide()
		except:
			if self.main.get_config().get_dict_query()==True:
				if (havepackagekit or haveaptfile):
					self.notifier.notify(_("Dictionary missing"), _("The spellcheck dictionary for '%(one)s' ('%(two)s') is not installed.")%{'one':lang[1], 'two':lang[2]}, ((_("Install"), self.__install_dictionary_start, lang[2]),(_("Don't ask again"), self.__disable_dictionary_autoinstall)))
				else:
					self.notifier.notify(_("Dictionary missing"), _("The spellcheck dictionary for '%(one)s' ('%(two)s') is not installed.")%{'one':lang[1], 'two':lang[2]}, ((_("Help"), lambda w,e: dialogs.help('Dictionaries')),(_("Don't show again"), self.__disable_dictionary_autoinstall)))
	
	def __disable_dictionary_autoinstall(self, w, e):
		""" Disables directory autoinstallation query. """
		self.notifier.hide()
		self.main.get_config().set_dict_query(False)
	
	def __install_dictionary_start(self, w, e, lang):
		""" Starts the dictionary installation. """
		self.notifier.hide()
		threadtools.SubprocessThread(self.__install_dictionary, self.__install_dictionary_end, lang).start()
		if haveaptfile:
			self.main.set_busy(True, _("Searching package..."))
	
	def __install_dictionary(self, lang):
		""" Performs dictionary installation. """
		if haveaptfile:
			p=subprocess.Popen(["apt-file","search","-x","^/usr/share/hunspell/%s.dic$"%lang], shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, creationflags=threadtools.popencf)
			out=p.communicate()
			if p.returncode!=0:
				return out[1]
			if out[0]=='':
				return _("No package providing the requested file found.")
			gobject.idle_add(self.main.pop_status)
			gobject.idle_add(self.main.push_status, _("Running transaction..."))
			package=out[0].split('\n')[0].split(':')[0]
			p=subprocess.Popen(["gksudo","apt-get -y install %s"%package], shell=False, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, creationflags=threadtools.popencf)
			out=p.communicate()
			if p.returncode!=0:
				if out[1]=='':
					return _("Transaction failed")
				return out[1]
		elif havepackagekit:
			try:
					pgkproxy=dbus.SessionBus().get_object('org.freedesktop.PackageKit', '/org/freedesktop/PackageKit')
					iface=dbus.Interface(pgkproxy, 'org.freedesktop.PackageKit.Modify')
					iface.InstallProvideFiles(dbus.UInt32(0), ["/usr/share/myspell/%s.dic"%lang, "/usr/share/hunspell/%s.dic"%lang], "always",timeout=300)
			except Exception, e:
				return str(e)
		return True
	
	def __install_dictionary_end(self, ret):
		""" Called when done installing the dictionary. """
		if haveaptfile:
			self.main.set_busy(False)
		if ret!=True:
			dialogs.error_dialog(_("Failed to install dictionary"),_("The following error occured:\n%s")%ret, self.window)
		else:
			self.main.load_dictionaries()
	
	def detach_spellchecker(self):
		""" Detaches the spellchecker, called from main on quit. """
		if self.spellchecker!=None:
			gtkspell.Spell.detach(self.spellchecker)
	
	### Buffer Functions ###
	def __append_mode_changed(self, widget):
		""" Called when the append mode was changed. """
		self.append_mode=widget.get_active()
	
	def __onselect(self,widget,data):
		""" Called when text is selected in the text view. """
		self.btn_stripcrlf.set_sensitive(self.outbuffer.get_has_selection()==True)
	
	def __stripcrlf(self, widget):
		""" Called when the strip newlines button is clicked. """
		self.output.set_editable(False)
		iters=self.outbuffer.get_selection_bounds()
		if len(iters)==0:
			return #should not happen
		txt=self.outbuffer.get_text(iters[0],iters[1])
		excl='';
		if self.menu_keepifdot.get_active(): excl+='\.'
		if self.menu_keepifquest.get_active(): excl+='\?'
		if self.menu_keepifexcl.get_active(): excl+='\!'
		if self.menu_keepifquot.get_active(): excl+='\'"'
		if excl!='':
			excl='([^'+excl+'])'
		else:
			excl='.'
		p=re.compile(excl+'\n')
		txt=p.sub(r'\1 ',txt)
		txt=txt.replace("  "," ")
		self.outbuffer.delete(iters[0],iters[1])
		self.outbuffer.insert_at_cursor(txt)
		self.output.set_editable(True)
		self.btn_stripcrlf.set_sensitive(False)
		
	def save_buffer(self, widget=None):
		""" Saves the buffer content. """
		savefile=dialogs.save_text(os.path.dirname(self.outputfile), os.path.basename(self.outputfile), self.window)
		if savefile=="":
			return False
		try:
			outfile = open(savefile, 'w')
			outfile.write(self.outbuffer.get_text(self.outbuffer.get_start_iter(),self.outbuffer.get_end_iter()))
			outfile.close()
			self.outbuffer.set_modified(False)
			return True
		except:
			dialogs.error_dialog(_("Failed to save output"),_("Check that you have writing permissions in the selected folder."),self.window)
			return False
	
	def clear_buffer(self, widget=None):
		""" Clears the buffer content. """
		if not (self.outputbox.flags() & gtk.VISIBLE):
			return True
		if self.outbuffer.get_modified():
			response=dialogs.question_dialog(_("Output not saved"),_("Save output before proceeding?"),self.window)
			if (response==1 and not self.save_buffer()) or response==2:
				return False
		self.outbuffer.set_text("")
		self.outputbox.hide()
		return True
	
	def set_buffer(self, text, outputfile):
		""" Sets the buffer text. """
		self.outputfile=outputfile
		if self.append_mode==0: #at cursor (selection)
			iters=self.outbuffer.get_selection_bounds()
			if len(iters)!=0:
				self.outbuffer.delete(iters[0],iters[1])			
			self.outbuffer.insert_at_cursor(text)
		elif self.append_mode==1: #at end
			self.outbuffer.insert(self.outbuffer.get_end_iter(),text)
		elif self.append_mode==2: #replace
			self.outbuffer.set_text(text)
		if self.spellchecker!=None:
			self.spellchecker.recheck_all()
		self.outputbox.show()

	### Find / Replace functions ###
	def __entry_find_changed(self, widget):
		""" Called when the find input changes. """
		self.entry_find.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse(self.main.get_config().get_color("base_color")))
		
	def __set_search_iter(self, widget, data):
		""" Sets the search iter. """
		self.searchiter=self.outbuffer.get_iter_at_mark(self.outbuffer.get_insert())
		
	def __toggle_findreplace(self, widget):
		""" Toggles the find-replace box. """
		if widget.get_active():
			self.findreplacebox.show()
			self.entry_find.grab_focus()
		else:
			self.findreplacebox.hide()
			self.entry_find.set_text('')
			self.entry_replace.set_text('')
			self.entry_find.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse(self.main.get_config().get_color("base_color")))
	
	def __find_in_buffer(self, widget, replace=False):
		""" Searches the search string in the buffer, and replaces any occurence if replace=True. """
		self.entry_find.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse(self.main.get_config().get_color("base_color")))
		find=self.entry_find.get_text()
		if find=="":
			return
		iters=self.outbuffer.get_selection_bounds()
		if len(iters)!=0:
			if self.outbuffer.get_text(iters[0], iters[1])==find:
				if replace:
					self.outbuffer.delete(iters[0],iters[1])
					self.outbuffer.insert_at_cursor(self.entry_replace.get_text())
				else:
					self.searchiter.forward_char()
		if self.searchiter==None or self.searchiter.is_end():
			self.searchiter=self.outbuffer.get_start_iter()
		try:
			self.searchiter, end = self.searchiter.forward_search(find, gtk.TEXT_SEARCH_VISIBLE_ONLY)
		except:
			try:
				self.searchiter, end = self.outbuffer.get_start_iter().forward_search(find, gtk.TEXT_SEARCH_VISIBLE_ONLY)
			except:
				self.entry_find.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse("#FF8080"))
				return
		self.outbuffer.select_range(self.searchiter,end)
		self.output.scroll_mark_onscreen(self.outbuffer.get_insert())
